<?php
namespace Ak\Form\Annotation;

use ArrayObject;
use Doctrine\Common\Annotations\Annotation;
use Doctrine\Common\Annotations\AnnotationReader;
use Doctrine\Common\Annotations\Reader as AnnotationReaderInterface;
use Laminas\EventManager\EventManagerInterface;
use Laminas\Form\Annotation\AnnotationBuilder as BaseAnnotationBuilder;
use Laminas\Form\Fieldset;
use Laminas\Stdlib\ArrayUtils;

/**
 * Class AnnotationBuilder
 *
 * @package Ak\Form\Annotation
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/autokreso/wp-core
 */
final class AnnotationBuilder extends BaseAnnotationBuilder
{
    /** @var AnnotationReaderInterface */
    private $reader;

    /**
     * AnnotationBuilder constructor.
     * @param AnnotationReaderInterface $reader
     * @throws \Doctrine\Common\Annotations\AnnotationException
     */
    public function __construct(AnnotationReaderInterface $reader = null)
    {
        $this->reader = $reader ?? new AnnotationReader();
    }

    /**
     * Set event manager instance
     *
     * @param  EventManagerInterface $events
     * @return AnnotationBuilder
     */
    public function setEventManager(EventManagerInterface $events)
    {
        $events->setIdentifiers([
            __CLASS__,
            get_class($this),
        ]);
        (new ElementAnnotationsListener())->attach($events);
        (new FormAnnotationsListener())->attach($events);
        $this->events = $events;
        return $this;
    }

    /**
     * Creates and returns a form specification for use with a factory
     *
     * Parses the object provided, and processes annotations for the class and
     * all properties. Information from annotations is then used to create
     * specifications for a form, its elements, and its input filter.
     *
     * @param  string|object $entity Either an instance or a valid class name for an entity
     * @return ArrayObject
     * @throws \ReflectionException
     */
    public function getFormSpecification($entity)
    {
        if (! is_object($entity)) {
            if ((is_string($entity) && (! class_exists($entity))) // non-existent class
                || (! is_string($entity)) // not an object or string
            ) {
                throw new \InvalidArgumentException(sprintf(
                    '%s expects an object or valid class name; received "%s"',
                    __METHOD__,
                    var_export($entity, 1)
                ));
            }
        }

        $this->entity      = $entity;
        $formSpec          = new ArrayObject();
        $filterSpec        = new ArrayObject();

        $reflection  = new \ReflectionClass($entity);
        $annotations = $this->getClassAnnotations($reflection);

        if (count($annotations) > 0) {
            $this->configureForm($annotations, $reflection, $formSpec, $filterSpec);
        }

        foreach ($reflection->getProperties() as $property) {
            $annotations = $this->getPropertyAnnotations($property);
            $this->configureElement($annotations, $property, $formSpec, $filterSpec);
        }

        if (! isset($formSpec['input_filter'])) {
            $formSpec['input_filter'] = $filterSpec;
        } elseif (is_array($formSpec['input_filter'])) {
            $formSpec['input_filter'] = ArrayUtils::merge($filterSpec->getArrayCopy(), $formSpec['input_filter']);
        }

        if (!isset($filterSpec['options'])) {
            $formSpec['options'] = [];
        }

        $formSpec['options']['allowed_object_binding_class'] = is_object($entity) ? get_class($entity) : $entity;

        return $formSpec;
    }

    /**
     * Configure the form specification from annotations
     *
     * @param  array $annotations
     * @param  \ReflectionClass $reflection
     * @param  ArrayObject $formSpec
     * @param  ArrayObject $filterSpec
     * @return void
     * @triggers discoverName
     * @triggers configureForm
     */
    protected function configureForm($annotations, $reflection, $formSpec, $filterSpec)
    {
        /** @noinspection PhpParamsInspection */
        $name                   = $this->discoverName($annotations, $reflection);
        $formSpec['name']       = $name;
        $formSpec['attributes'] = [];
        $formSpec['elements']   = [];
        $formSpec['fieldsets']  = [];

        $events = $this->getEventManager();

        foreach ($annotations as $annotation) {
            $events->trigger(__FUNCTION__, $this, [
                'annotation' => $annotation,
                'name'       => $name,
                'formSpec'   => $formSpec,
                'filterSpec' => $filterSpec,
            ]);
        }
    }

    /**
     * Configure an element from annotations
     *
     * @param  array $annotations
     * @param  \ReflectionProperty $reflection
     * @param  ArrayObject $formSpec
     * @param  ArrayObject $filterSpec
     * @return void
     * @triggers checkForExclude
     * @triggers discoverName
     * @triggers configureElement
     * @throws \ReflectionException
     */
    protected function configureElement($annotations, $reflection, $formSpec, $filterSpec)
    {
        // If the element is marked as exclude, return early
        /** @noinspection PhpParamsInspection */
        if ($this->checkForExclude($annotations)) {
            return;
        }

        $events = $this->getEventManager();
        /** @noinspection PhpParamsInspection */
        $name   = $this->discoverName($annotations, $reflection);

        $elementSpec = new ArrayObject([
            'flags' => [],
            'spec'  => [
                'name' => $name
            ],
        ]);

        $inputSpec = new ArrayObject([
            'name' => $name,
        ]);

        $params = new ArrayObject([
            'name'        => $name,
            'elementSpec' => $elementSpec,
            'inputSpec'   => $inputSpec,
            'formSpec'    => $formSpec,
            'filterSpec'  => $filterSpec,
        ]);

        foreach ($annotations as $annotation) {
            $params['annotation'] = $annotation;
            $events->trigger(__FUNCTION__, $this, $params);
        }

        // Since "type" is a reserved name in the filter specification,
        // we need to add the specification without the name as the key.
        // In all other cases, though, the name is fine.
        if ($params['inputSpec']->count() > 1) {
            if ($name === 'type') {
                $filterSpec[] = $params['inputSpec'];
            } else {
                $filterSpec[$name] = $params['inputSpec'];
            }
        }

        $elementSpec = $params['elementSpec'];
        $type        = (isset($elementSpec['spec']['type']))
            ? $elementSpec['spec']['type']
            : 'Laminas\Form\Element';

        if (class_exists($type) && !is_subclass_of($type, 'Laminas\Form\ElementInterface')) {
            $options = $elementSpec['spec']['options'] ?? [];
            $elementSpec['spec'] = $this->getFormSpecification($type);
            $elementSpec['spec']['name'] = $name;
            $elementSpec['spec']['options'] = $options;
            $elementSpec['spec']['type'] = Fieldset::class;

            if (! isset($formSpec['elements'])) {
                $formSpec['elements'] = [];
            }

            $formSpec['elements'][] = $elementSpec;

        } else {
            // Compose as a fieldset or an element, based on specification type.
            // If preserve defined order is true, all elements are composed as elements to keep their ordering
            if (! $this->preserveDefinedOrder() && is_subclass_of($type, 'Laminas\Form\FieldsetInterface')) {
                if (! isset($formSpec['fieldsets'])) {
                    $formSpec['fieldsets'] = [];
                }
                $formSpec['fieldsets'][] = $elementSpec;
            } else {
                if (! isset($formSpec['elements'])) {
                    $formSpec['elements'] = [];
                }
                $formSpec['elements'][] = $elementSpec;
            }
        }
    }

    /**
     * Vraća sve anotacije postavljene na klasu, uključujući i parent klase.
     *
     * @param \ReflectionClass $class
     * @return Annotation[]
     */
    private function getClassAnnotations(\ReflectionClass $class): array
    {
        $classes = $this->getClassTree($class);
        $annotations = [];

        foreach ($classes as $class) {
            $annotations = ArrayUtils::merge($annotations, $this->reader->getClassAnnotations($class));
        }

        return $annotations;
    }

    /**
     * @param \ReflectionProperty $property
     * @return Annotation[]
     */
    private function getPropertyAnnotations(\ReflectionProperty $property): array
    {
        $properties = $this->getPropertyTree($property);
        $annotations = [];

        foreach ($properties as $property) {
            $annotations = ArrayUtils::merge($annotations, $this->reader->getPropertyAnnotations($property));
        }

        return $annotations;
    }

    /**
     * @param \ReflectionClass $class
     * @return \ReflectionClass[]
     */
    private function getClassTree(\ReflectionClass $class): array
    {
        $classes = [$class];

        while ($parent = $class->getParentClass()) {
            array_unshift($classes, $parent);
            $class = $class->getParentClass();
        }

        return $classes;
    }

    /**
     * @param \ReflectionProperty $property
     * @return \ReflectionProperty[]
     */
    private function getPropertyTree(\ReflectionProperty $property): array
    {
        $properties = [];
        $classes = $this->getClassTree($property->getDeclaringClass());

        foreach ($classes as $class) {
            if ($class->hasProperty($property->getName())) {
                $properties[] = $class->getProperty($property->getName());
            }
        }

        return $properties;
    }
}
